""" Define helper methods to reduce the boilerplate code for a server
"""

import logging

from fastapi import Request
from fastapi.exceptions import RequestValidationError
from fastapi.openapi.utils import get_openapi
from fastapi.responses import JSONResponse
from starlette.exceptions import HTTPException as StarletteHTTPException
from styler_rest_framework import message
from styler_rest_framework.config import defaults
from styler_rest_framework.middlewares.fastapi import exception_middleware
from styler_rest_framework.middlewares.fastapi.auth_middleware import add_auth_middleware
from styler_rest_framework.logging.error_reporting import google_error_reporting_handler


def api_error_reporting_handler(service=None):  # pragma: no coverage
    try:
        service_name = service or defaults.ERROR_HANDLER_SERVICE
        from google.cloud import error_reporting

        handler = google_error_reporting_handler(service=service_name)

        def error_handler(request, exc):
            http_context = error_reporting.HTTPContext(
                method=request.method, url=request.url.path
            )
            handler(http_context=http_context)

        return error_handler
    except Exception:
        logging.warning("Could not find start api error reporting")
        return None


def add_middlewares(app, error_handler=None, service=None, handle_exceptions_args=None):
    """Append default exception middleware

    Args:
        app: FastAPI app
        error_handler: exception handler
        service: service name
        handle_exceptions_args: dict of parameters to the middleware
    """
    handle_exceptions_args = handle_exceptions_args or {}
    error_handler = error_handler or api_error_reporting_handler(service=service)

    exception_middleware.add_exception_middleware(
        app, error_handler=error_handler, **handle_exceptions_args
    )


def setup_validation_handler(
    app, validation_error_code=422, validation_code="validation_error"
):  # pragma: no coverage
    """Set the validation error by overriding the RequestValidationError"""

    @app.exception_handler(RequestValidationError)
    async def validation_exception_handler(
        request: Request, exc: RequestValidationError
    ):
        locale = request.headers.get("Accept-Language", "ja")
        errors = {}
        for err in exc.errors():
            field = ".".join(err["loc"][1:]) or "error"
            err_type = err["type"].replace(".", "_")
            errors[field] = message.get(
                f"pydantic_validation.{err_type}", locale=locale, **err.get("ctx", {})
            )

        return JSONResponse(
            status_code=validation_error_code,
            content={"code": validation_code, "reason": errors},
        )

    def custom_openapi():
        if app.openapi_schema:
            return app.openapi_schema
        openapi_schema = get_openapi(
            title=defaults.SERVICE_NAME, version=defaults.VERSION, routes=app.routes
        )
        openapi_schema["components"]["schemas"].pop("ValidationError")
        openapi_schema["components"]["schemas"]["HTTPValidationError"] = {
            "title": "ValidationError",
            "type": "object",
            "properties": {
                "code": {"type": "string"},
                "reason": {
                    "type": "object",
                    "additionalProperties": {"type": "string"},
                },
            },
            "required": ["code", "reason"],
        }
        app.openapi_schema = openapi_schema
        return app.openapi_schema

    app.openapi = custom_openapi


def setup_standard_error_format(app):  # pragma: no coverage
    """Set the validation error by overriding the HTTPException"""

    @app.exception_handler(StarletteHTTPException)
    async def http_exception_handler(r: Request, exc: StarletteHTTPException):
        return JSONResponse(
            status_code=exc.status_code,
            content=exc.detail,
        )



""" Tests for the SendMail message
"""

import json

from styler_rest_framework.pubsub.publishers.messages.send_mail_message \
    import (
        SendMailMessage
    )


class TestSendMailMessage:
    """ Tests for SendMail message
    """
    def test_generate_message(self):
        msg = SendMailMessage(
            'mail@mail.com',
            'Some title',
            'Body here'
        )
        expected_keys_in_message = {
            'sender',
            'receiver',
            'cc',
            'title',
            'body',
            'type',
        }

        message_data = msg.encoded()
        decoded = json.loads(str(message_data, encoding='utf-8'))

        assert 'name' in decoded
        assert decoded['name'] == 'SendMail'
        assert 'arg' in decoded
        assert all([k in decoded['arg'] for k in expected_keys_in_message])



""" Publishing module
"""

from styler_rest_framework.pubsub.publishers.pubsub_handler import publish_message


def publisher_for_message(message_type, topic):
    """Returns a callable to handle publishing a message"""

    def wrapper(*args, **kwargs):
        msg = message_type(*args, **kwargs)
        publish_message(topic, msg.encoded())

    return wrapper



""" Error reporting module
"""
import logging

from styler_rest_framework.config import defaults


def google_error_reporting_handler(service=None):  # pragma: no coverage
    try:
        service_name = service or defaults.ERROR_HANDLER_SERVICE
        from google.cloud import error_reporting

        client = error_reporting.Client(service=service_name)
        return client.report_exception

    except Exception:
        logging.warning("Could not find start error reporting client")
        return None



from typing import Dict, List


class Datasource:
    def all_documents(self, db, path: str):
        query = db.collection(path)
        return [doc.to_dict() for doc in query.stream()]

    def save_document(self, db, path: str, doc, doc_id):
        return db.document(f"{path}/{doc_id}").set(doc)

    def query_with_arguments(
        self, db, path: str, where: List = None, order_by: Dict = None
    ):
        """Perform simple and compound queries in Firestore

        :param db: Firestore client
        :type db: Firestore client
        :param path: path to the collection
        :type path: str
        :param where: A list of where object, which is a Dict must contain the following content:
            * field_path (str) – A field path (.-delimited list of field names) for the field to filter on.
            * op_string (str) – A comparison operation in the form of a string. Acceptable values are <, <=, ==, !=, >=, >, in, not-in, array_contains and array_contains_any.
            * value (Any) – The value to compare the field against in the filter. If value is None or a NaN, then == is the only allowed operation.
            Query for firestore has some limitations, please check https://firebase.google.com/docs/firestore/query-data/queries#query_limitations for more information.
        :type where: List, optional
        :param order_by: A Dict of order_by object, which should contain the following content:
            * field_path (str) – A field path (.-delimited list of field names) on which to order the query results.
            * direction (Optional[str]) – The direction to order by. Must be one of ASCENDING or DESCENDING, defaults to ASCENDING.
        :type order_by: Dict, optional
        """
        query = db.collection(path)
        if where:
            if not isinstance(where, list):
                raise ValueError("where argument should be a list!")
            for obj in where:
                if "field_path" in obj and "op_string" in obj and "value" in obj:
                    query = query.where(
                        obj["field_path"], obj["op_string"], obj["value"]
                    )
                else:
                    raise ValueError(
                        'where object must contain "field_path", "op_string" and "value" fields!'
                    )
        if order_by:
            if not isinstance(order_by, dict):
                raise ValueError("order_by argument should be a dictionary!")
            if "field_path" in order_by and "direction" in order_by:
                if order_by["direction"] not in ("ASCENDING", "DESCENDING"):
                    raise ValueError(
                        'direction of order_by must be "ASCENDING" or "DESCENDING"!'
                    )
                else:
                    query = query.order_by(
                        order_by["field_path"], order_by["direction"]
                    )
            elif "field_path" in order_by:
                query = query.order_by(order_by["field_path"])
            else:
                raise ValueError("invalid order_by argument!")
        return [doc.to_dict() for doc in query.stream()]




from styler_rest_framework.api.authorization import highest_role, authorize, Role


class TestHighestRole:
    def test_highest_role(self):
        roles = []
        assert highest_role(roles) == 0
        roles.append('staff')
        assert highest_role(roles) == 1
        roles.append('admin')
        assert highest_role(roles) == 2
        roles.append('sysadmin')
        assert highest_role(roles) == 4


class TestAuthorize:
    def test_return_decorator(self):
        assert callable(authorize(Role.ADMIN))


class TestDecorator:
    def test_minimum_role(self, token):

        @authorize(Role.ADMIN)
        def my_endpoint(authorization):
            return None

        tk = token(sysadmin=False, admin=False, staff=True)

        response = my_endpoint(authorization=f'Bearer {tk}')
        assert response.status_code == 403

        tk = token(sysadmin=False, admin=True, staff=True)
        response = my_endpoint(authorization=f'Bearer {tk}')
        assert response is None

        tk = token(sysadmin=True, admin=False, staff=True)
        response = my_endpoint(authorization=f'Bearer {tk}')
        assert response is None

    async def test_minimum_role_no_token(self, token):

        @authorize(Role.ADMIN)
        def my_endpoint():
            return None

        response = my_endpoint()
        assert response.status_code == 403

    async def test_minimum_role_async(self, token):

        @authorize(Role.ADMIN)
        async def my_endpoint(authorization):
            return None

        tk = token(sysadmin=False, admin=False, staff=True)

        response = await my_endpoint(authorization=f'Bearer {tk}')
        assert response.status_code == 403

        tk = token(sysadmin=False, admin=True, staff=True)
        response = await my_endpoint(authorization=f'Bearer {tk}')
        assert response is None

        tk = token(sysadmin=True, admin=False, staff=True)
        response = await my_endpoint(authorization=f'Bearer {tk}')
        assert response is None

    async def test_minimum_role_async_no_token(self, token):

        @authorize(Role.ADMIN)
        async def my_endpoint():
            return None

        response = await my_endpoint()
        assert response.status_code == 403



"""Tests for GCSHandler
"""
from unittest.mock import MagicMock, patch
import pytest

from google.api_core.exceptions import ServiceUnavailable

from styler_rest_framework.helpers.gcs_handler import GCSHandler


class TestGetFilesFromFolder:
    """Tests for function get_files_from_folder"""

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_normal_flow(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_bucket.list_blobs.return_value = "test return"

        test_handler = GCSHandler("test_bucket_name")
        result = test_handler.get_files_from_folder("test_folder")
        assert result == "test return"
        mock_bucket.list_blobs.assert_called_once()

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_succeed(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_bucket.list_blobs.side_effect = [
            ServiceUnavailable("service unavailable"),
            "test return",
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        result = test_handler.get_files_from_folder("test_folder")
        assert result == "test return"
        assert mock_bucket.list_blobs.call_count == 2

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_failed(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_bucket.list_blobs.side_effect = [
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        with pytest.raises(ServiceUnavailable) as expected:
            test_handler.get_files_from_folder("test_folder")
            assert expected.value.errors == "service unavailable"


class TestDownloadFileAsStringWithFormatter:
    """Tests for function download_file_as_string_with_formatter"""

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_normal_flow(self, mock_gcs, csv_blob_from_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.download_as_string.return_value = csv_blob_from_gcs

        test_handler = GCSHandler("test_bucket_name")
        result = test_handler.download_file_as_string_with_formatter("test_file_path")
        assert len(result) == 2
        assert result[0] == "header1,header2"
        assert result[1] == "row1-1,row1-2"
        mock_bucket.blob.assert_called_once()
        mock_blob.download_as_string.assert_called_once()

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_succeed(self, mock_gcs, csv_blob_from_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.download_as_string.side_effect = [
            ServiceUnavailable("service unavailable"),
            csv_blob_from_gcs,
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        result = test_handler.download_file_as_string_with_formatter("test_file_path")
        assert len(result) == 2
        assert result[0] == "header1,header2"
        assert result[1] == "row1-1,row1-2"
        assert mock_bucket.blob.call_count == 2
        assert mock_blob.download_as_string.call_count == 2

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_failed(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.download_as_string.side_effect = [
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        with pytest.raises(ServiceUnavailable) as expected:
            test_handler.download_file_as_string_with_formatter("test_file_path")
            assert expected.value.errors == "service unavailable"


class TestDownloadFileAsStream:
    """Tests for function download_file_as_stream"""

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_normal_flow(self, mock_gcs, csv_blob_from_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.download_as_bytes.return_value = csv_blob_from_gcs

        test_handler = GCSHandler("test_bucket_name")
        result = test_handler.download_file_as_stream("test_file_path")
        assert result.getvalue() == csv_blob_from_gcs.decode("utf-8")
        mock_bucket.blob.assert_called_once()
        mock_blob.download_as_bytes.assert_called_once()

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_succeed(self, mock_gcs, csv_blob_from_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.download_as_bytes.side_effect = [
            ServiceUnavailable("service unavailable"),
            csv_blob_from_gcs,
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        result = test_handler.download_file_as_stream("test_file_path")
        assert result.getvalue() == csv_blob_from_gcs.decode("utf-8")
        assert mock_bucket.blob.call_count == 2
        assert mock_blob.download_as_bytes.call_count == 2

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_failed(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.download_as_bytes.side_effect = [
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        with pytest.raises(ServiceUnavailable) as expected:
            test_handler.download_file_as_stream("test_file_path")
            assert expected.value.errors == "service unavailable"


class TestDownloadFileAsString:
    """Tests for function download_file_as_string"""

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_normal_flow(self, mock_gcs, csv_blob_from_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.download_as_string.return_value = csv_blob_from_gcs

        test_handler = GCSHandler("test_bucket_name")
        result = test_handler.download_file_as_string("test_file_path")
        assert result == csv_blob_from_gcs
        mock_bucket.blob.assert_called_once()
        mock_blob.download_as_string.assert_called_once()

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_succeed(self, mock_gcs, csv_blob_from_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.download_as_string.side_effect = [
            ServiceUnavailable("service unavailable"),
            csv_blob_from_gcs,
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        result = test_handler.download_file_as_string("test_file_path")
        assert result == csv_blob_from_gcs
        assert mock_bucket.blob.call_count == 2
        assert mock_blob.download_as_string.call_count == 2

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_failed(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.download_as_string.side_effect = [
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        with pytest.raises(ServiceUnavailable) as expected:
            test_handler.download_file_as_string("test_file_path")
            assert expected.value.errors == "service unavailable"


class TestUploadFile:
    """Tests for function upload_file"""

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_normal_flow(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.upload_from_file.return_value = True

        test_handler = GCSHandler("test_bucket_name")
        result = test_handler.upload_file("test_file_obj", "test_file_path")
        assert result is True
        mock_bucket.blob.assert_called_once()
        mock_blob.upload_from_file.assert_called_once()

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_succeed(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.upload_from_file.side_effect = [
            ServiceUnavailable("service unavailable"),
            True,
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        result = test_handler.upload_file("test_file_obj", "test_file_path")
        assert result is True
        assert mock_bucket.blob.call_count == 2
        assert mock_blob.upload_from_file.call_count == 2

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_failed(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.upload_from_file.side_effect = [
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        with pytest.raises(ServiceUnavailable) as expected:
            test_handler.upload_file("test_file_obj", "test_file_path")
            assert expected.value.errors == "service unavailable"


class TestRenameFile:
    """Tests for function rename_file"""

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_normal_flow(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_bucket.rename_blob.return_value = True

        test_handler = GCSHandler("test_bucket_name")
        result = test_handler.rename_file("old_file_name", "new_file_name")
        assert result is True
        mock_bucket.blob.assert_called_once()
        mock_bucket.rename_blob.assert_called_once()

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_succeed(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_bucket.rename_blob.side_effect = [
            ServiceUnavailable("service unavailable"),
            True,
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        result = test_handler.rename_file("old_file_name", "new_file_name")
        assert result is True
        assert mock_bucket.blob.call_count == 2
        assert mock_bucket.rename_blob.call_count == 2

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_failed(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_bucket.rename_blob.side_effect = [
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        with pytest.raises(ServiceUnavailable) as expected:
            test_handler.rename_file("old_file_name", "new_file_name")
            assert expected.value.errors == "service unavailable"


class TestPutAsString:
    """Tests for function put_as_string"""

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_normal_flow(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.upload_from_string.return_value = True

        test_handler = GCSHandler("test_bucket_name")
        result = test_handler.put_as_string(
            "test_file_name", "test_file_string", "test_content_type"
        )
        assert result is True
        mock_bucket.blob.assert_called_once()
        mock_blob.upload_from_string.assert_called_once()

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_succeed(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.upload_from_string.side_effect = [
            ServiceUnavailable("service unavailable"),
            True,
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        result = test_handler.put_as_string(
            "test_file_name", "test_file_string", "test_content_type"
        )
        assert result is True
        assert mock_bucket.blob.call_count == 2
        assert mock_blob.upload_from_string.call_count == 2

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_failed(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.upload_from_string.side_effect = [
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        with pytest.raises(ServiceUnavailable) as expected:
            test_handler.put_as_string(
                "test_file_name", "test_file_string", "test_content_type"
            )
            assert expected.value.errors == "service unavailable"


class TestDeleteFile:
    """Tests for function delete_file"""

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_normal_flow(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.delete.return_value = True

        test_handler = GCSHandler("test_bucket_name")
        result = test_handler.delete_file("test_file_name")
        assert result is True
        mock_bucket.blob.assert_called_once()
        mock_blob.delete.assert_called_once()

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_succeed(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.delete.side_effect = [ServiceUnavailable("service unavailable"), True]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        result = test_handler.delete_file("test_file_name")
        assert result is True
        assert mock_bucket.blob.call_count == 2
        assert mock_blob.delete.call_count == 2

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_failed(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.delete.side_effect = [
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        with pytest.raises(ServiceUnavailable) as expected:
            test_handler.delete_file("test_file_name")
            assert expected.value.errors == "service unavailable"


class TestIsExist:
    """Tests for function is_exist"""

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_normal_flow(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.exists.return_value = True

        test_handler = GCSHandler("test_bucket_name")
        result = test_handler.is_exist("test_file_name")
        assert result is True
        mock_bucket.blob.assert_called_once()
        mock_blob.exists.assert_called_once()

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_succeed(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.exists.side_effect = [ServiceUnavailable("service unavailable"), True]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        result = test_handler.is_exist("test_file_name")
        assert result is True
        assert mock_bucket.blob.call_count == 2
        assert mock_blob.exists.call_count == 2

    @patch("styler_rest_framework.helpers.gcs_handler.storage.Client", autospec=True)
    def test_retry_failed(self, mock_gcs):
        mock_client = mock_gcs.return_value
        mock_client.get_bucket = MagicMock()
        mock_bucket = mock_client.get_bucket.return_value
        mock_blob = mock_bucket.blob.return_value
        mock_blob.exists.side_effect = [
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
            ServiceUnavailable("service unavailable"),
        ]

        test_handler = GCSHandler("test_bucket_name", back_off=0.1)
        with pytest.raises(ServiceUnavailable) as expected:
            test_handler.is_exist("test_file_name")
            assert expected.value.errors == "service unavailable"




from styler_rest_framework.crypto import aes
import pytest


class TestEncrypt:
    """ Tests for encrypt method
    """
    def test_no_key(self):
        with pytest.raises(ValueError) as expected:
            _ = aes.encrypt(None, 'something')

        assert str(expected.value) == 'Key is missing'

    def test_no_data(self):
        with pytest.raises(ValueError) as expected:
            _ = aes.encrypt('key', None)

        assert str(expected.value) == 'Data is missing'

    def test_encrypt_data(self):
        result = aes.encrypt('a long key', 'some data')

        decrypt = aes.decrypt('a long key', result)

        assert decrypt == 'some data'


class TestDecrypt:
    """ Tests for decrypt method
    """
    def test_no_key(self):
        with pytest.raises(ValueError) as expected:
            _ = aes.decrypt(None, 'data')

        assert str(expected.value) == 'Key is missing'

    def test_no_data(self):
        with pytest.raises(ValueError) as expected:
            _ = aes.decrypt('key', None)

        assert str(expected.value) == 'Data is missing'

    def test_decrypt_data(self):
        result = aes.encrypt('a long key', 'some data')
        result = '1234' + result[4:]

        with pytest.raises(aes.DecryptError):
            _ = aes.decrypt('a long key', result)

    def test_decrypt_modified_data(self):
        result = aes.encrypt('a long key', 'some data')

        decrypt = aes.decrypt('a long key', result)

        assert decrypt == 'some data'



import logging
import time
from random import randint

from alembic.config import Config
from alembic import command


class MigrationError(Exception):
    """MigrationError class for exceptions in this module."""

    pass


def get_current_version(alembic_cfg):
    captured_text = None

    def print_stdout(text, *arg):
        nonlocal captured_text
        captured_text = text

    alembic_cfg.print_stdout = print_stdout
    command.current(alembic_cfg)
    return captured_text


def get_heads_version(alembic_cfg):
    captured_text = []

    def print_stdout(text, *arg):
        nonlocal captured_text
        captured_text.append(text)

    alembic_cfg.print_stdout = print_stdout
    command.heads(alembic_cfg)
    return captured_text


def retry_migration(alembic_cfg, count, max_count, max_backoff=32):
    try:
        logging.warning(f"retry migration now... (retry count:{count+1})")
        command.upgrade(alembic_cfg, "head")
        logging.warning("retry migration succeed!")
        return
    except Exception:
        backoff = min(2 ** (count) + randint(1, 1000) / 1000, max_backoff)
        count += 1
        if count >= max_count:
            raise MigrationError("Could not execute alembic migrations.")
        logging.warning(
            f"retry migration failed, \
            will retry again after {backoff} seconds..."
        )
        time.sleep(backoff)
        retry_migration(alembic_cfg, count, max_count, max_backoff)


def check_and_retry_migration(cfg_path, max_retry_count=3):
    if not cfg_path:
        raise MigrationError("parameter required: cfg_path")
    alembic_cfg = Config(cfg_path)
    heads = get_heads_version(alembic_cfg)
    if len(heads) > 1:
        raise MigrationError(f"multiple heads detected! {heads}")
    current = get_current_version(alembic_cfg)
    if current == heads[0]:
        return
    logging.warning(
        "alembic version is not up-to-date, \
        will retry migration..."
    )
    retry_migration(alembic_cfg, 0, max_retry_count)




from functools import wraps
import asyncio

from fastapi.responses import JSONResponse
from styler_rest_framework.api.request_scope import RequestScope


class Role:
    SYSADMIN = 4
    ADMIN = 2
    STAFF = 1


def highest_role(roles):
    if 'sysadmin' in roles:
        return Role.SYSADMIN
    elif 'admin' in roles:
        return Role.ADMIN
    elif 'staff' in roles:
        return Role.STAFF
    else:
        return 0


def authorize(role: int):
    """Verifies if the JWT contains at least the minimum required role.

    :param role: Minimum required role
    :type role: int
    """
    def has_roles(func):
        @wraps(func)
        async def awrapper(*args, **kwargs):
            token = kwargs.get('authorization')
            if not token:
                return JSONResponse(status_code=403)
            identity = RequestScope(authorization=token)
            if highest_role(identity._roles()) < role:
                return JSONResponse(status_code=403)
            return await func(*args, **kwargs)

        @wraps(func)
        def wrapper(*args, **kwargs):
            token = kwargs.get('authorization')
            if not token:
                return JSONResponse(status_code=403)
            identity = RequestScope(authorization=token)
            if highest_role(identity._roles()) < role:
                return JSONResponse(status_code=403)
            return func(*args, **kwargs)

        if asyncio.iscoroutinefunction(func):
            return awrapper
        else:
            return wrapper

    return has_roles



""" Request scope for FastAPI
"""

from typing import Dict, Callable

from jwt.exceptions import InvalidTokenError
import jwt


class Identity:
    """Holds the identity of the logged user"""

    def __init__(self, token):
        self._token = token
        try:
            self._decoded = jwt.decode(self._token, options={"verify_signature": False})
        except InvalidTokenError:
            raise ValueError("Invalid JWT token")

    def user_id(self):
        """Returns the user_id provided by firebase"""
        return self._decoded.get("user_id")

    def is_system_admin(self):
        """Returns a boolean identifying the user as a system administrator"""
        return "sysadmin" in self._roles()

    def is_admin(self):
        """Returns a boolean identifying the user
        as an organization administrator
        """
        return "admin" in self._roles()

    def is_staff(self):
        """Returns a boolean identifying the user
        as a shop staff
        """
        return "staff" in self._roles()

    def shops(self):
        """Returns a list of shop_ids that the user has access to"""
        return self._custom_claims().get("shop", [])

    def organizations(self):
        """Returns a list of organization_ids that the user has access to"""
        return self._custom_claims().get("organization", [])

    def data(self):
        """Return the entire data from the token"""
        return self._decoded

    def token(self):
        """Returns the original JWT token"""
        return self._token

    def _roles(self):
        """Returns the collection of roles"""
        if "roles" not in self._decoded:
            return []
        return self._decoded["roles"]

    def _custom_claims(self):
        """Returns the collection of custom claims"""
        if "claims" not in self._decoded:
            return {}
        return self._decoded["claims"]


class RequestScope(Identity):
    """Carries information about the request scope"""

    def __init__(
        self,
        authorization: str = None,
        accept_language: str = "ja",
        trace: Dict = None,
    ) -> Callable:
        self.accept_language = accept_language
        self.trace = trace or {}
        token = get_token(authorization)
        super().__init__(token)

    def localization(self):
        # Also 'ja' when None is passed.
        return self.accept_language or "ja"

    def trace_header(self):
        return self.trace

    @classmethod
    def from_request(cls, request):
        auth_header = request.headers.get("Authorization")
        accept_language = request.headers.get("Accept-Language", "ja")
        trace = request.headers.get("trace_header", {})
        return cls(
            authorization=auth_header, accept_language=accept_language, trace=trace
        )


def get_token(authorization: str) -> str:
    """Obtains the Access Token from the Authorization Header value"""
    if not authorization:
        raise ValueError("Authorization header is expected")

    parts = authorization.split()

    if parts[0].lower() != "bearer":
        raise ValueError("Authorization header must start with Bearer")
    elif len(parts) == 1:
        raise ValueError("Token not found")
    elif len(parts) > 2:
        raise ValueError("Authorization header must be Bearer token")

    token = parts[1]
    return token



from datetime import datetime
from time import time
import logging

try:
    from cryptography.x509 import load_pem_x509_certificate
    from cryptography.hazmat.backends import default_backend
    from jwt.exceptions import InvalidTokenError
    import jwt
    import requests
except Exception:
    logging.error('Missing libraries:  try  pipenv install "pyjwt[crypto]"')


CURRENT_JWKS = None
EXPIRATION = None


def retrieve_kid(token):  # pragma: no coverage
    return jwt.get_unverified_header(token)['kid']


def public_key(x509):  # pragma: no coverage
    cert_obj = load_pem_x509_certificate(str.encode(x509), default_backend())
    return cert_obj.public_key();


def google_x509(url):  # pragma: no coverage
    response = requests.get(url)
    return response.json(), datetime.strptime(response.headers['Expires'], '%a, %d %b %Y %H:%M:%S GMT')


def verify_token(tk_json):  # pragma: no coverage
    if tk_json['exp'] < time():
        raise ValueError('Expired')
    if tk_json['iat'] > time():
        raise ValueError('Invalid issued at time')
    if tk_json['auth_time'] > time():
        raise ValueError('Invalid authentication time')

def validate(token, env, jwks_url):  # pragma: no coverage
    global CURRENT_JWKS
    global EXPIRATION

    # Verify if the keys have expired
    if not EXPIRATION or EXPIRATION < datetime.utcnow():
        # Refresh JWKS
        CURRENT_JWKS, EXPIRATION = google_x509(jwks_url)
    try:
        # Retrieve the kid from the token
        kid = retrieve_kid(token)

        # Retrieve the public key from the x509 certificate
        pub = public_key(CURRENT_JWKS[kid])

        # Validate JWT using the public key

        data = jwt.decode(
            token,
            pub,
            algorithms=["RS256"],
            audience=f'facy-{env}',
            issuer=f'https://securetoken.google.com/facy-{env}',
            options={'verify_exp': True}
        )
        verify_token(data)
        return data
    except InvalidTokenError as ex:
        logging.exception(ex)
        return None



""" Conftest
"""

import pytest
import jwt

from styler_rest_framework.datasource import firestore


@pytest.fixture
def event_loop(loop):
    return loop


@pytest.fixture
def token():
    def generate(
        overwrites=None,
        sysadmin=False,
        admin=False,
        staff=False,
        shops=None,
        organizations=None,
        user_id=None,
    ):
        if overwrites is None:
            overwrites = {}
        data = {
            **{
                "roles": [],
                "claims": {"shop": [], "organization": []},
                "iss": "issuer",
                "aud": "audition",
                "auth_time": "time",
                "user_id": user_id or "1234",
                "sub": "sub",
                "iat": 1595838390,
                "exp": 1595839390,
                "email": "email@test.com",
                "email_verified": False,
                "firebase": {
                    "identities": {"email": ["email@test.com"]},
                    "sign_in_provider": "custom",
                },
            },
            **overwrites,
        }
        if sysadmin:
            data["roles"].append("sysadmin")
        if admin:
            data["roles"].append("admin")
        if staff:
            data["roles"].append("staff")
        if shops:
            data["claims"]["shop"].extend(shops)
            data["roles"].append("staff")
        if organizations:
            data["claims"]["organization"].extend(organizations)
            data["roles"].append("admin")
        return jwt.encode(data, "secret-key")

    return generate


@pytest.fixture
def empty_token():
    return jwt.encode({}, "secret-key")


@pytest.fixture
def csv_blob_from_gcs():
    return "header1,header2\nrow1-1,row1-2".encode("utf-8")


@pytest.fixture
def mock_db():
    """Mock firestore Client"""
    yield firestore.CLIENT


@pytest.fixture
def with_firestore_docs(mock_db):
    def generate(path, doc_count=1, id_prefix="test_document"):
        output = []
        for i in range(doc_count):
            doc_dict = {
                "number": i+1,
                "string": f"string{i}",
                "bool": i % 2 == 0,
                "nest_dict": {
                    "number": i+1,
                    "string": f"string{i}",
                    "bool": i % 2 == 0,
                },
                "list": [i, i + 1, i + 2],
            }
            output.append(doc_dict)
            mock_db.document(f"{path}/{id_prefix}{i}").set(doc_dict)
        return output

    return generate



""" Tests for request_scope module
"""

from unittest.mock import Mock

from styler_rest_framework.api.request_scope import (
    Identity,
    RequestScope,
    get_token,
)
import pytest


class TestIdentity:
    """ Tests
    """
    def test_invalid_token(self):
        with pytest.raises(ValueError) as expected:
            Identity('invalid token')

        assert str(expected.value) == 'Invalid JWT token'

    def test_valid_token(self, token):
        idem = Identity(token())

        assert isinstance(idem, Identity)

    def test_token(self, token):
        original_token = token()
        idem = Identity(original_token)

        tk = idem.token()

        assert original_token == tk

    def test_user_id(self, token):
        idem = Identity(token({'user_id': '1234'}))

        user_id = idem.user_id()

        assert user_id == '1234'

    def test_is_system_admin(self, token):
        idem = Identity(token(sysadmin=True))
        idem2 = Identity(token(admin=True))

        it_should_be = idem.is_system_admin()
        is_should_not_be = idem2.is_system_admin()

        assert it_should_be
        assert not is_should_not_be

    def test_is_admin(self, token):
        idem = Identity(token(admin=True))
        idem2 = Identity(token(staff=True))

        it_should_be = idem.is_admin()
        is_should_not_be = idem2.is_admin()

        assert it_should_be
        assert not is_should_not_be

    def test_is_staff(self, token):
        idem = Identity(token(staff=True))
        idem2 = Identity(token(admin=True))

        it_should_be = idem.is_staff()
        is_should_not_be = idem2.is_staff()

        assert it_should_be
        assert not is_should_not_be

    def test_invalid_roles_sysadmin(self, empty_token):
        idem = Identity(empty_token)

        result = idem.is_system_admin()

        assert not result

    def test_invalid_roles_admin(self, empty_token):
        idem = Identity(empty_token)

        result = idem.is_admin()

        assert not result

    def test_invalid_roles_staff(self, empty_token):
        idem = Identity(empty_token)

        result = idem.is_staff()

        assert not result

    def test_shops(self, token):
        idem = Identity(token(shops=['12345', '33442']))

        shops = idem.shops()

        assert shops == ['12345', '33442']

    def test_shops_none(self, empty_token):
        idem = Identity(empty_token)

        shops = idem.shops()

        assert shops == []

    def test_organizations(self, token):
        idem = Identity(token(organizations=['33333']))

        organizations = idem.organizations()

        assert organizations == ['33333']

    def test_organizations_none(self, empty_token):
        idem = Identity(empty_token)

        organizations = idem.organizations()

        assert organizations == []

    def test_data(self, token):
        idem = Identity(token())
        expected_keys = {
            'claims',
            'roles',
            'iss',
            'aud',
            'auth_time',
            'user_id',
            'sub',
            'iat',
            'exp',
            'email',
            'email_verified',
            'firebase'
        }

        data = idem.data()

        assert set(data.keys()) == expected_keys


class TestGetToken:
    """ Tests for get_token
    """
    def test_get_token(self):
        """ Should return the token
        """
        token = get_token('Bearer TOKEN')

        assert token == 'TOKEN'

    def test_none(self):
        """ Should raise an exception
        """
        with pytest.raises(ValueError) as expected:
            get_token(None)
        assert str(expected.value) == 'Authorization header is expected'

    def test_without_bearer(self):
        """ Should raise an exception if there is no bearer
        """
        with pytest.raises(ValueError) as expected:
            get_token('token TOKEN')
        assert str(expected.value) == \
            'Authorization header must start with Bearer'

    def test_without_token(self):
        """ Should raise an exception if the auth is invalid
        """
        with pytest.raises(ValueError) as expected:
            get_token('Bearer')
        assert str(expected.value) == 'Token not found'

    def test_invalid_auth(self):
        """ Should raise an exception if the Authorization is invalid
        """
        with pytest.raises(ValueError) as expected:
            get_token('Bearer TOKEN TOKEN')
        assert str(expected.value) == \
            'Authorization header must be Bearer token'


@pytest.fixture
def auth():
    return (
        'Bearer '
        'eyJhbGciOiJSUzI1NiIsImtpZCI6Im'
        'MzZjI3NjU0MmJmZmU0NWU5OG'
        'MyMGQ2MDNlYmUyYmExMTc2ZWRhMzMiLCJ0e'
        'XAiOiJKV1QifQ.eyJzeXN0ZW1fYWRta'
        'W4iOmZhbHNlLCJjbGFpbXMiOnsic2hvcHMi'
        'OlsiN2YwNzZjNDgtMGU0Yy00MDQ0LTl'
        'kNDctMjM0MjM0MzI0MjM0IiwiN2YwNzZjND'
        'gtMGU0Yy00MDQ0LTlkNDctMzg4OTk5M'
        'Dk4MDkwIl0sIm9yZ2FuaXphdGlvbnMiOltd'
        'fSwiaXNzIjoiaHR0cHM6Ly9zZWN1cmV'
        '0b2tlbi5nb29nbGUuY29tL2ZhY3ktZGV2ZW'
        'xvcG1lbnQiLCJhdWQiOiJmYWN5LWRld'
        'mVsb3BtZW50IiwiYXV0aF90aW1lIjoxNTky'
        'ODg2Mjg5LCJ1c2VyX2lkIjoiNm5hZjg'
        'wanR5YlZSdVAyMmVZY0lZR1gyS0pLMiIsIn'
        'N1YiI6IjZuYWY4MGp0eWJWUnVQMjJlW'
        'WNJWUdYMktKSzIiLCJpYXQiOjE1OTI4ODYy'
        'ODksImV4cCI6MTU5Mjg4OTg4OSwiZW1'
        'haWwiOiJicnVuby5zdWdhbm9Ac3R5bGVyLm'
        'xpbmsiLCJlbWFpbF92ZXJpZmllZCI6Z'
        'mFsc2UsImZpcmViYXNlIjp7ImlkZW50aXRp'
        'ZXMiOnsiZW1haWwiOlsiYnJ1bm8uc3V'
        'nYW5vQHN0eWxlci5saW5rIl19LCJzaWduX2'
        'luX3Byb3ZpZGVyIjoiY3VzdG9tIn19.'
        'VNJtVQ9nmvbpUkMM95cdN06O3NDUqdBwEgm'
        'czfWKR2Vt5fOFQGtFykEovEmajdVHZO'
        'm4tbJlbfoILu0i_GsO-7jaoYNBOC9-tIBiD'
        '8pvUCeL4iQj2jhpvR6z3PEtTyA7i7Ae'
        'VUj9wuxrzYm0wtcLTR1vSEp4jTj3Nw5KD4a'
        'dPq4IF63SKqywNLkmgURuQnF-WlaZBn'
        'g2CK0AMddY0Qt-1Z0IFUp-i_tXkGMOPBjXs'
        'wq1kPxyZ2xtNWZjQZ5sWNw2dmkC2wyf'
        'EmnQ5Zm9UJvrS6tiCigTGkF43PhwTl26K9z'
        'twOtekrpOiwap-c55rUG8TbobpmGGkF'
        'i9Q3r5dHH-ag'
    )


class TestLocalization:
    """ Tests for localization
    """
    def test_localization(self, auth):
        req_scope = RequestScope(
            authorization=auth,
            accept_language='en'
        )

        localization = req_scope.localization()

        assert localization == 'en'

    def test_localization_none(self, auth):
        req_scope = RequestScope(
            authorization=auth
        )

        localization = req_scope.localization()

        assert localization == 'ja'


class TestTraceHeader:
    """ Tests for trace_header
    """
    def test_trace_header(self, auth):
        req_scope = RequestScope(
            authorization=auth,
            trace={'aa': 'bb'}
        )

        trace_header = req_scope.trace_header()

        assert trace_header == {'aa': 'bb'}

    def test_trace_header_none(self, auth):
        req_scope = RequestScope(
            authorization=auth
        )

        trace_header = req_scope.trace_header()

        assert trace_header == {}


class MockHeaders:
    def __init__(self, auth, locale, header=None):
        self.locale = locale
        self.header = header
        self.auth = auth

    def get(self, key, default=None):
        if key == 'Authorization':
            return self.auth
        if key == 'Accept-Language':
            return self.locale or default
        if key == 'trace_header':
            return self.header or {}


class TestFromRequest:
    """ Tests for class method from_request
    """
    def test_from_request(self, auth):
        request = Mock()
        request.headers = MockHeaders(
            auth=auth, locale='en', header={'aa': '1234'})

        req_scope = RequestScope.from_request(request)

        assert req_scope.localization() == 'en'
        assert req_scope.trace_header() == {'aa': '1234'}



import logging

from aiohttp import web
from opencensus.common.transports.async_ import AsyncTransport
from opencensus.ext.stackdriver import trace_exporter as stackdriver_exporter
from opencensus.trace import attributes_helper
from opencensus.trace.propagation import google_cloud_format
import opencensus.trace.tracer


HTTP_HOST = attributes_helper.COMMON_ATTRIBUTES["HTTP_HOST"]
HTTP_METHOD = attributes_helper.COMMON_ATTRIBUTES["HTTP_METHOD"]
HTTP_PATH = attributes_helper.COMMON_ATTRIBUTES["HTTP_PATH"]
HTTP_ROUTE = attributes_helper.COMMON_ATTRIBUTES["HTTP_ROUTE"]
HTTP_URL = attributes_helper.COMMON_ATTRIBUTES["HTTP_URL"]
HTTP_STATUS_CODE = attributes_helper.COMMON_ATTRIBUTES["HTTP_STATUS_CODE"]


def initialize_tracer(project_id, context, propagator):
    exporter = stackdriver_exporter.StackdriverExporter(
        project_id=project_id, transport=AsyncTransport
    )
    tracer = opencensus.trace.tracer.Tracer(
        span_context=context,
        exporter=exporter,
        propagator=propagator,
        sampler=opencensus.trace.tracer.samplers.AlwaysOnSampler(),
    )

    return tracer


def trace_middleware(project_id):
    @web.middleware
    async def middleware(request, handler):
        if handler.__name__ == "health_check":
            return await handler(request)
        span = None
        tracer = None
        try:
            propagator = google_cloud_format.GoogleCloudFormatPropagator()
            span_context = propagator.from_headers(request.headers)
            tracer = initialize_tracer(project_id, span_context, propagator)
            span = tracer.start_span()
            span.name = handler.__name__
            tracer.add_attribute_to_current_span(HTTP_HOST, request.host)
            tracer.add_attribute_to_current_span(HTTP_METHOD, request.method)
            tracer.add_attribute_to_current_span(HTTP_PATH, request.path)
            tracer.add_attribute_to_current_span(HTTP_URL, str(request.url))
            request["trace_header"] = propagator.to_headers(span_context)
        except:  # NOQA
            logging.exception("Could not initialize the tracer")

        try:
            response = await handler(request)
            if tracer:
                tracer.add_attribute_to_current_span(HTTP_STATUS_CODE, response.status)
            return response
        finally:
            if tracer:
                tracer.end_span()

    return middleware


def config_tracer(app, project_id):
    app.middlewares.append(trace_middleware(project_id))



""" Tests for the LogMe message
"""

import json

from styler_rest_framework.pubsub.publishers.messages.logme_message \
    import (
        LogMeMessage
    )


class TestLogMeMessage:
    """ Tests for SendMail message
    """
    def test_generate_message(self):
        msg = LogMeMessage(
            'test_dataset',
            'test_table',
            [{"field": "value"}, {"field": "value2"}]
        )
        expected_keys_in_message = {
            'dataset',
            'table',
            'rows'
        }

        message_data = msg.encoded()
        decoded = json.loads(str(message_data, encoding='utf-8'))

        assert 'name' in decoded
        assert decoded['name'] == 'LogMe'
        assert 'arg' in decoded
        assert all([k in decoded['arg'] for k in expected_keys_in_message])


"""Tests for Firestore Handler
"""
import pytest

from styler_rest_framework.datasource import Datasource

DATA_ROOT_PATH = "firestore-test/v1/data"


class TestSaveDocument:
    """Tests for function save_document"""

    def test_normal_flow(self, mock_db):
        """Tests normal flow for function save_document"""
        mock_db.reset()
        test_data = Datasource()
        test_doc = {"test_key": "test_value"}
        test_doc_id = "test_doc_1"
        test_data.save_document(
            mock_db,
            DATA_ROOT_PATH,
            test_doc,
            test_doc_id,
        )

        doc_content = (
            mock_db.document(f"{DATA_ROOT_PATH}/{test_doc_id}").get().to_dict()
        )

        assert doc_content == test_doc


class TestSaveDocument:
    """Tests for function all_documents"""

    def test_normal_flow(self, mock_db, with_firestore_docs):
        """Tests normal flow for function save_document"""
        mock_db.reset()
        doc_count = 10
        test_docs = with_firestore_docs(DATA_ROOT_PATH, doc_count=doc_count)
        test_data = Datasource()
        result = test_data.all_documents(mock_db, DATA_ROOT_PATH)
        assert len(result) == doc_count
        for i, doc in enumerate(result):
            assert doc == test_docs[i]


class TestQueryWithArguments:
    """Tests for function query_with_arguments"""

    def test_single_where_object(self, mock_db, with_firestore_docs):
        """Tests normal flow for function query_with_arguments"""
        mock_db.reset()
        doc_count = 10
        test_docs = with_firestore_docs(DATA_ROOT_PATH, doc_count=doc_count)
        test_data = Datasource()
        where_list = [{"field_path": "number", "op_string": ">", "value": 5}]
        result = test_data.query_with_arguments(
            mock_db, DATA_ROOT_PATH, where=where_list
        )
        assert len(result) == 5
        for i, doc in enumerate(result):
            assert doc == test_docs[i + 5]

    def test_multiple_where_object(self, mock_db, with_firestore_docs):
        """Tests normal flow for function query_with_arguments"""
        mock_db.reset()
        doc_count = 10
        test_docs = with_firestore_docs(DATA_ROOT_PATH, doc_count=doc_count)
        test_data = Datasource()
        where_list = [
            {"field_path": "number", "op_string": ">", "value": 5},
            {"field_path": "number", "op_string": "<", "value": 8},
        ]
        result = test_data.query_with_arguments(
            mock_db, DATA_ROOT_PATH, where=where_list
        )
        assert len(result) == 2
        assert result[0] == test_docs[5]
        assert result[1] == test_docs[6]

    def test_invalid_where_object_type(self, mock_db, with_firestore_docs):
        """Tests normal flow for function query_with_arguments"""
        mock_db.reset()
        doc_count = 10
        test_docs = with_firestore_docs(DATA_ROOT_PATH, doc_count=doc_count)
        test_data = Datasource()
        where_list = {"field_path": "number", "op_string": ">", "value": 5}
        with pytest.raises(ValueError) as expected:
            test_data.query_with_arguments(mock_db, DATA_ROOT_PATH, where=where_list)
            assert expected.value.errors == "where argument should be a list!"

    def test_invalid_key_in_where_object(self, mock_db, with_firestore_docs):
        """Tests normal flow for function query_with_arguments"""
        mock_db.reset()
        doc_count = 10
        test_docs = with_firestore_docs(DATA_ROOT_PATH, doc_count=doc_count)
        test_data = Datasource()
        where_list = [{"invalid_key": "number"}]
        with pytest.raises(ValueError) as expected:
            test_data.query_with_arguments(mock_db, DATA_ROOT_PATH, where=where_list)
            assert (
                expected.value.errors
                == 'where object must contain "field_path", "op_string" and "value" fields!'
            )

    def test_with_order_by_alone(self, mock_db, with_firestore_docs):
        """Tests normal flow for function query_with_arguments"""
        mock_db.reset()
        doc_count = 10
        test_docs = with_firestore_docs(DATA_ROOT_PATH, doc_count=doc_count)
        test_data = Datasource()
        order_by = {"field_path": "number", "direction": "DESCENDING"}
        result = test_data.query_with_arguments(
            mock_db, DATA_ROOT_PATH, order_by=order_by
        )
        assert len(result) == 10
        assert result[0] == test_docs[-1]
        assert result[-1] == test_docs[0]

    def test_with_both_where_and_order_by(self, mock_db, with_firestore_docs):
        """Tests normal flow for function query_with_arguments"""
        mock_db.reset()
        doc_count = 10
        test_docs = with_firestore_docs(DATA_ROOT_PATH, doc_count=doc_count)
        test_data = Datasource()
        where_list = [
            {"field_path": "nest_dict.number", "op_string": ">", "value": 5},
            {"field_path": "nest_dict.number", "op_string": "<", "value": 8},
        ]
        order_by = {"field_path": "number"}
        result = test_data.query_with_arguments(
            mock_db, DATA_ROOT_PATH, where=where_list, order_by=order_by
        )
        assert len(result) == 2
        assert result[0] == test_docs[5]
        assert result[1] == test_docs[6]

    def test_invalid_order_by_type(self, mock_db, with_firestore_docs):
        """Tests normal flow for function query_with_arguments"""
        mock_db.reset()
        doc_count = 10
        test_docs = with_firestore_docs(DATA_ROOT_PATH, doc_count=doc_count)
        test_data = Datasource()
        order_by = [{"field_path": "number", "direction": "DESCENDING"}]

        with pytest.raises(ValueError) as expected:
            test_data.query_with_arguments(mock_db, DATA_ROOT_PATH, order_by=order_by)
            assert expected.value.errors == "order_by argument should be a dictionary!"

    def test_invalid_order_by_key(self, mock_db, with_firestore_docs):
        """Tests normal flow for function query_with_arguments"""
        mock_db.reset()
        doc_count = 10
        test_docs = with_firestore_docs(DATA_ROOT_PATH, doc_count=doc_count)
        test_data = Datasource()
        order_by = {"direction": "DESCENDING"}

        with pytest.raises(ValueError) as expected:
            test_data.query_with_arguments(mock_db, DATA_ROOT_PATH, order_by=order_by)
            assert expected.value.errors == "invalid order_by argument!"

    def test_invalid_direction_in_order_by(self, mock_db, with_firestore_docs):
        """Tests normal flow for function query_with_arguments"""
        mock_db.reset()
        doc_count = 10
        test_docs = with_firestore_docs(DATA_ROOT_PATH, doc_count=doc_count)
        test_data = Datasource()
        order_by = [{"field_path": "number", "direction": "invalid"}]

        with pytest.raises(ValueError) as expected:
            test_data.query_with_arguments(mock_db, DATA_ROOT_PATH, order_by=order_by)
            assert expected.value.errors == 'direction of order_by must be "ASCENDING" or "DESCENDING"!'



# -*- coding: utf-8 -*-
"""This is the class of handling google cloud storage"""

from io import StringIO
import time

from google.cloud import storage
from google.api_core.exceptions import ServiceUnavailable


class GCSHandler:
    """The handler of google cloud storage (GCS)

    Args:
        bucket_name (str): target bucket name of gs
        *args: placeholder
        **kwargs: placeholder

    Attributes:
        bucket : a gs bucket instance
    """

    def __init__(self, bucket_name, *args, **kwargs):
        client = storage.Client()
        self.bucket = client.get_bucket(bucket_name)
        self._BACKOFF = kwargs.get("back_off", 3)

    def get_files_from_folder(self, folder_name, retry=3):
        """Return an iterator used to find blobs in the bucket.

        Args:
            folder_name (str): folder name or prefix path

        Returns:
            an iterator used to find blobs in the bucket
        """
        try:
            return self.bucket.list_blobs(prefix=folder_name)
        except ServiceUnavailable:
            if retry > 0:
                time.sleep(self._BACKOFF)
                return self.get_files_from_folder(folder_name, retry=retry - 1)
            else:
                raise

    def download_file_as_string_with_formatter(self, file_path, retry=3):
        """Download the contents of this blob as a bytes object.

        Args:
            file_path (str): path of file

        Returns:
            The data stored in this blob.
        """
        try:
            blob = self.bucket.blob(file_path)
            return blob.download_as_string().decode("utf-8").splitlines()
        except ServiceUnavailable:
            if retry > 0:
                time.sleep(self._BACKOFF)
                return self.download_file_as_string_with_formatter(
                    file_path, retry=retry - 1
                )
            else:
                raise

    def download_file_as_stream(self, file_path, retry=3):
        """Download the contents of this blob as a StringIO object.

        Args:
            file_path (str): path of file

        Returns:
            The data stored in this blob.
        """
        try:
            blob = self.bucket.blob(file_path)
            return StringIO(blob.download_as_bytes().decode("utf-8"))
        except ServiceUnavailable:
            if retry > 0:
                time.sleep(self._BACKOFF)
                return self.download_file_as_stream(file_path, retry=retry - 1)
            else:
                raise

    def download_file_as_string(self, file_path, retry=3):
        """Download the contents of this blob as a bytes object.

        Args:
            file_path (str): path of file

        Returns:
            The data stored in this blob.
        """
        try:
            blob = self.bucket.blob(file_path)
            return blob.download_as_string()
        except ServiceUnavailable:
            if retry > 0:
                time.sleep(self._BACKOFF)
                return self.download_file_as_string(file_path, retry=retry - 1)
            else:
                raise

    def upload_file(self, file_obj, file_name, retry=3):
        """upload a file to the bucket.

        Args:
            file_obj (file object): A file handle open for reading.
            file_name (str): file name

        Raises:
            GoogleCloudError if the upload response returns an error status.
        """
        try:
            blob = self.bucket.blob(file_name)
            return blob.upload_from_file(file_obj)
        except ServiceUnavailable:
            if retry > 0:
                time.sleep(self._BACKOFF)
                return self.upload_file(file_obj, file_name, retry=retry - 1)
            else:
                raise

    def rename_file(self, old_file_name, new_file_name, retry=3):
        """rename a file from the bucket.

        Args:
            old_file_name (str): old file name
            new_file_name (str): new file name

        Raises:
            GoogleCloudError if the upload response returns an error status.
        """
        try:
            blob = self.bucket.blob(old_file_name)
            return self.bucket.rename_blob(blob, new_file_name)
        except ServiceUnavailable:
            if retry > 0:
                time.sleep(self._BACKOFF)
                return self.rename_file(old_file_name, new_file_name, retry=retry - 1)
            else:
                raise

    def put_as_string(
        self, filename: str, file_string: bytes, content_type: str, retry=3
    ):
        """Method to upload objects."""
        try:
            blob = self.bucket.blob(filename)
            return blob.upload_from_string(file_string, content_type)
        except ServiceUnavailable:
            if retry > 0:
                time.sleep(self._BACKOFF)
                return self.put_as_string(
                    filename, file_string, content_type, retry=retry - 1
                )
            else:
                raise

    def delete_file(self, filename, retry=3):
        """Deletes a file from the bucket"""
        try:
            blob = self.bucket.blob(filename)
            return blob.delete()
        except ServiceUnavailable:
            if retry > 0:
                time.sleep(self._BACKOFF)
                return self.delete_file(filename, retry=retry - 1)
            else:
                raise

    def is_exist(self, filename, retry=3):
        """Check if a file exists in the bucket"""
        try:
            blob = self.bucket.blob(filename)
            return blob.exists()
        except ServiceUnavailable:
            if retry > 0:
                time.sleep(self._BACKOFF)
                return self.is_exist(filename, retry=retry - 1)
            else:
                raise



""" Tests for the shops publisher
"""

from unittest.mock import Mock, patch

from styler_rest_framework.pubsub.publishers import (
    pubsub_handler as handler,
    publisher_for_message,
)
from styler_rest_framework.pubsub.publishers.messages import (
    Message
)
import pytest


class TestGetCallback:
    """ Tests for get_callback
    """
    def test_callback(self):
        """ Should return a callable
        """
        data = Mock()

        result = handler.get_callback(data)

        assert callable(result)

    @patch('logging.error')
    def test_callback_raise(self, mocked_logging):
        """ The callable should raise an error
        """
        data = Mock()
        result = handler.get_callback(data)

        with pytest.raises(Exception) as expected:
            api_future = Mock()
            api_future.result.side_effect = Exception('Something went wrong')
            result(api_future)

        assert str(expected.value) == 'Something went wrong'
        mocked_logging.assert_called_once()


class MockMessage(Message):
    def __init__(self, some_data):
        self.name = 'MockMessage'
        self.arg = some_data


class TestPublisher:
    """Tests for publisher
    """
    @patch('styler_rest_framework.pubsub.publishers.publish_message')
    def test_publisher_for_message(self, mocked_publisher):
        func = publisher_for_message(MockMessage, 'my-topic')

        func({'aaa': 123})

        mocked_publisher.assert_called_once()



""" Helper to build responses
"""
from typing import Dict
import logging

from styler_rest_framework.schemas.error import (
    HTTP400Error,
    HTTP401Error,
    HTTP402Error,
    HTTP403Error,
    HTTP404Error,
    HTTP409Error,
)
from fastapi import HTTPException
from styler_rest_framework.exceptions.business import (
    PermissionDeniedError,
    ResourceNotFoundError,
    ValidationError,
    ConflictError,
)
from styler_rest_framework.exceptions.services import (
    AuthenticationError,
    AuthorizationError,
    InvalidDataError,
    NotFoundError,
    PaymentRequiredError,
    ConflictError as ServiceConflictError,
)


def standard(*codes) -> Dict:
    """Generate the default configuration for error codes"""
    codes = codes or (400, 401, 402, 403, 404, 409)
    responses = {}
    if 400 in codes:
        responses[400] = {"model": HTTP400Error}
    if 401 in codes:
        responses[401] = {"model": HTTP401Error}
    if 402 in codes:
        responses[402] = {"model": HTTP402Error}
    if 403 in codes:
        responses[403] = {"model": HTTP403Error}
    if 404 in codes:
        responses[404] = {"model": HTTP404Error}
    if 409 in codes:
        responses[409] = {"model": HTTP409Error}
    return responses


def bad_request(errors: Dict[str, str], code="validation_error"):
    """Return HTTP 400 error"""
    raise HTTPException(
        status_code=400,
        detail={
            "code": code,
            "reason": errors,
        },
    )


def payment_required(code: str, reason: str):
    """Return HTTP 402 error"""
    raise HTTPException(
        status_code=402,
        detail={
            "code": code,
            "reason": reason,
        },
    )


def forbidden(msg=None):
    """Return HTTP 403 error"""
    msg = msg or "Permission denied"
    raise HTTPException(status_code=403, detail=msg)


def not_found(msg=None):
    """Return HTTP 404 error"""
    msg = msg or "Resource not found"
    raise HTTPException(status_code=404, detail=msg)


def unauthorized(msg=None):
    """Return HTTP 401 error"""
    msg = msg or "Invalid credentials"
    raise HTTPException(status_code=401, detail=msg)


def conflict(msg=None):
    """Return HTTP 409 error"""
    msg = msg or "Resource conflict"
    raise HTTPException(status_code=409, detail=msg)


def internal_server_error(exception, msg="An error has occurred"):
    """Return HTTP 500 error"""
    logging.exception("Internal Server error: %s", str(exception))
    raise HTTPException(status_code=500, detail=msg)


def handle_business_errors(exception: Exception) -> None:
    """Handle standard business exceptions"""
    if isinstance(exception, PermissionDeniedError):
        forbidden()
    elif isinstance(exception, ResourceNotFoundError):
        not_found()
    elif isinstance(exception, ValidationError):
        bad_request(exception.errors)
    elif isinstance(exception, ConflictError):
        conflict(exception.msg)
    else:
        internal_server_error(exception)


def handle_service_errors(ex):
    """Default method for handling service related errors"""
    if isinstance(ex, InvalidDataError):
        bad_request(ex.json_body().get("reason"))
    elif isinstance(ex, AuthenticationError):
        unauthorized()
    elif isinstance(ex, PaymentRequiredError):
        body = ex.json_body()
        payment_required(body.get("code"), body.get("reason"))
    elif isinstance(ex, AuthorizationError):
        forbidden()
    elif isinstance(ex, NotFoundError):
        not_found()
    elif isinstance(ex, ServiceConflictError):
        conflict()
    else:
        internal_server_error(exception=ex)



""" Message handling module
"""

from typing import Any, Callable, Dict
import json
import logging

from styler_rest_framework.logging import error_reporting
from styler_rest_framework.config import defaults


class MessageRouter:
    """Handles and routes messages to mapping handlers"""

    def __init__(self, error_handler=None):
        self._routes = {}
        self.error_handler = (
            error_handler
            or error_reporting.google_error_reporting_handler(
                service=defaults.ERROR_HANDLER_SERVICE
            )
        )

    def handle_message(self, message: Any) -> None:
        """Route the incoming messages

        Args:
            message: message received from pull
        """
        logging.info("Received message: %s", message)

        try:
            body = self._validate_envelope(message.data)

            message_name = body["name"]
            if message_name in self._routes:
                self._routes[message_name](body["arg"])
            else:
                logging.info("Unknown message name: %s", message_name)
            message.ack()
        except Exception:
            if self.error_handler:
                self.error_handler()
            logging.exception("Error when handling the message")

    def add_route(
        self, message_name: str, message_handler: Callable[[Any, Any], None]
    ) -> None:
        """Register a message route

        Args:
            message_name: expected message name.
            message_handler: function that receives a database connection
                and the args received in the message.
        """
        self._routes[message_name] = message_handler

    def _validate_envelope(self, data: str) -> (bool, Dict):
        """Validates the envelope of the data

        Args:
            data: the data received from the message
        Returns:
            The JSON parsed body data
        """
        body = json.loads(data)
        if "name" not in body or "arg" not in body:
            raise KeyError("Missing name or arg in the message body")
        return body



""" Handles i18n
"""

import os

from i18n import t
import i18n


dir_path = os.path.dirname(os.path.realpath(__file__))
i18n.set("enable_memoization", True)
i18n.load_path.append(f"{dir_path}/labels")
i18n.set("locale", "ja")
i18n.set("fallback", "ja")


def get(key, locale="ja", **kwargs):  # pragma: no coverage
    return t(key, locale=locale, **kwargs)


def load_path(path):  # pragma: no coverage
    i18n.load_path.append(path)



""" Handles creation of messages
"""

import json


class Message:
    """Base message"""

    def encoded(self):
        """Return a encoded data for pub sub"""
        json_string = json.dumps(self, default=lambda o: o.__dict__, indent=4)
        return json_string.encode("utf-8")



""" Request scope handles the identification of the user and request options
"""

import warnings

from styler_rest_framework.api.request_scope import Identity


class RequestScope(Identity):
    """Carries information about the request scope. (deprecated)"""

    def __init__(self, request):
        warnings.warn(
            """
            This class is deprecated and will be removed in a future release.
            Use RequestScope from styler_rest_framework.api.request_scope
            module instead.
            """
        )
        self.request = request
        token = get_token_auth_header(request)
        super().__init__(token)

    def localization(self):
        return self.request.headers.get("Accept-Language", "ja")

    def trace_header(self):
        return self.request.headers.get("trace_header", {})


def get_token_auth_header(request):
    """Obtains the Access Token from the Authorization Header"""
    auth_header = request.headers.get("Authorization")
    if not auth_header:
        raise ValueError("Authorization header is expected")

    parts = auth_header.split()

    if parts[0].lower() != "bearer":
        raise ValueError("Authorization header must start with Bearer")
    elif len(parts) == 1:
        raise ValueError("Token not found")
    elif len(parts) > 2:
        raise ValueError("Authorization header must be Bearer token")

    token = parts[1]
    return token



""" Paginator
"""

from math import ceil


class InvalidParameterError(ValueError):
    """Raised when page or limit is invalid"""


class Paginator:
    """Handles pagination"""

    def __init__(self, limit=20, page=1):
        try:
            if not page:
                page = 1
            self._current = int(page)
        except ValueError:
            raise InvalidParameterError("Invalid page")

        try:
            if not limit:
                limit = 20
            self._limit = int(limit)
        except ValueError:
            raise InvalidParameterError("Invalid limit")

        if self._current < 1:
            raise InvalidParameterError("Invalid page")
        if self._limit < 1:
            raise InvalidParameterError("Invalid limit")

        self._count = None
        self._total_pages = None

    def get_page(self, query):
        """Retrieve a page of items

        Args:
            query: query object
        Returns:
            A list of items within the boundaries set by the pagination
        """
        self._count = query.count()
        self._total_pages = int(ceil(self._count / float(self._limit)))
        offset = (self._current - 1) * self._limit
        if self._total_pages > 0 and self._current > self._total_pages:
            raise InvalidParameterError("Invalid page")
        return query.offset(offset).limit(self._limit).all()

    def get_info(self):
        """Retrieves information about the pagination status"""
        nex_page = self._current + 1 if self._current < self._total_pages else None
        return {
            "total_pages": self._total_pages,
            "total_number_of_items": self._count,
            "current_page": self._current,
            "previous_page": self._current - 1 if self._current > 1 else None,
            "next_page": nex_page,
        }



""" Tests for user_event module
"""
from datetime import datetime
from decimal import Decimal
from unittest.mock import patch, Mock
import json

from styler_rest_framework.events import user_event
from styler_rest_framework.api.request_scope import Identity


class TestGetUserId:
    """ Tests for user_event.get_user_id
    """
    def test_get_user_id(self, token):
        data = {
            0: Identity(token(sysadmin=True, user_id='111')),
            1: 'some parameter'
        }

        result = user_event.get_user_id(data)

        assert result == '111'

    def test_no_identity(self, token):
        data = {
            0: '1234',
            1: 'some parameter'
        }

        result = user_event.get_user_id(data)

        assert result is None

    def test_no_data(self, token):
        result = user_event.get_user_id(None)

        assert result is None


class MyCustomClass:
    def __init__(self):
        self.prop1 = '1234'
        self.prop2 = 5555


class TestExtractNamedArgs:
    """ Tests for user_event.extract_named_args
    """
    def test_extract_named_args(self):
        now = datetime.now()
        kwargs = {
            'key1': 'string',
            'key2': 1234,
            'key3': Decimal('12.30'),
            'key4': now,
            'key5': MyCustomClass()
        }

        result = user_event.extract_named_args(kwargs)

        assert result == {
            'key1': 'string',
            'key2': 1234,
            'key3': repr(Decimal('12.30')),
            'key4': repr(now),
            'key5': '{"prop1": "1234", "prop2": 5555}'
        }

    def test_no_data(self):
        result = user_event.extract_named_args(None)

        assert result == {}


class TestExtractArgs:
    """ Tests for user_event.extract_args
    """
    def test_extract_args(self):
        now = datetime.now()
        args = [1, 'something', now, Decimal('20.01'), MyCustomClass()]

        result = user_event.extract_args(args)

        assert result == {
            0: 1,
            1: 'something',
            2: repr(now),
            3: repr(Decimal('20.01')),
            4: '{"prop1": "1234", "prop2": 5555}'
        }


class TestTrack:
    """ Tests for user_event.track
    """
    @patch('styler_rest_framework.events.user_event.handler')
    def test_track_function(self, mocked_handler, token):
        @user_event.track
        def do_something(a, b, c, d):
            return 'something'
        iden = Identity(token(sysadmin=True, user_id='111'))

        with patch(
            'styler_rest_framework.events.user_event.time.time',
            Mock(return_value='timestamp')
        ):
            do_something(1234, 'aaa', c='parameter c', d=iden)

        mocked_handler.assert_called_once_with(
            user_event._TABLE,
            'tests.events.test_user_event.do_something',
            '111#tests.events.test_user_event.do_something#timestamp',
            {
                0: 1234,
                1: 'aaa',
                'c': 'parameter c',
                'd': json.dumps(iden.__dict__),
                'return': '"something"'
            }
        )



""" Defaults used by all modules
"""

import os


SERVICE_NAME = os.getenv("SERVICE_NAME") or "svc"
CONTAINER_NAME = os.getenv("CONTAINER_NAME") or "api"
NAMESPACE = os.getenv("NAMESPACE") or "namespace"
VERSION = os.getenv("VERSION") or "0.1.0"

ERROR_HANDLER_SERVICE = f"{NAMESPACE}/{CONTAINER_NAME}"

ENVIRONMENT = os.getenv("ENVIRONMENT")
TOPIC_NAME = os.getenv("TOPIC_NAME") or f"projects/facy-{ENVIRONMENT}/topics/common"
MAILER_TOPIC = os.getenv("MAILER_TOPIC") or f"projects/facy-{ENVIRONMENT}/topics/send-email"
LOGME_TOPIC = os.getenv("LOGME_TOPIC") or f"projects/facy-{ENVIRONMENT}/topics/logme"

# Mailer
EMAIL_SENDER = os.getenv("EMAIL_SENDER") or "info@styler.link"
EMAIL_SENDER_NAME = os.getenv("EMAIL_SENDER_NAME") or "FACYカスタマーサポート"
EMAIL_TYPE = os.getenv("EMAIL_TYPE") or "text/html"

# JWKS
JWKS_URL = os.getenv("JWKS_URL") or "https://www.googleapis.com/robot/v1/metadata/x509/securetoken@system.gserviceaccount.com"



""" Tests for the exception middleware
"""
from unittest.mock import Mock, patch, AsyncMock

from styler_rest_framework.middlewares.fastapi import auth_middleware
from fastapi import HTTPException
from fastapi.responses import JSONResponse
import pytest


class MockFastAPI:
    def __init__(self):
        self.middleware_func = None

    def middleware(self, middleware_type):
        def decorator(func):
            self.middleware_func = func
        return decorator


class TestAddAuthMiddleware:
    """ Tests for add_auth_middleware
    """
    def test_add_auth_middleware(self):
        app = MockFastAPI()

        auth_middleware.add_auth_middleware(app, 'development')

        assert app.middleware_func is not None
        assert callable(app.middleware_func)


class TestValidateJWTException:
    """ Tests for validating JWT
    """
    @patch('styler_rest_framework.middlewares.fastapi.auth_middleware.validate', Mock(return_value=True))
    async def test_valid_jwt(self):
        app = MockFastAPI()
        auth_middleware.add_auth_middleware(app, 'development')
        call_next = AsyncMock()
        request = Mock()
        request.headers.get.return_value = 'Bearer some_jwt'

        _ = await app.middleware_func(request, call_next)

        call_next.assert_called_once()

    @patch('styler_rest_framework.middlewares.fastapi.auth_middleware.validate', Mock(return_value=True))
    async def test_missing_jwt(self):
        app = MockFastAPI()
        auth_middleware.add_auth_middleware(app, 'development')
        call_next = AsyncMock()
        request = Mock()
        request.headers.get.return_value = None

        response = await app.middleware_func(request, call_next)

        call_next.assert_not_called()
        assert isinstance(response, JSONResponse)
        assert response.status_code == 401

    @patch('styler_rest_framework.middlewares.fastapi.auth_middleware.validate', Mock(return_value=False))
    async def test_invalid_jwt(self):
        app = MockFastAPI()
        auth_middleware.add_auth_middleware(app, 'development')
        call_next = AsyncMock()
        request = Mock()
        request.headers.get.return_value = 'Bearer some_jwt'
        
        response = await app.middleware_func(request, call_next)

        call_next.assert_not_called()
        assert isinstance(response, JSONResponse)
        assert response.status_code == 401



""" SaveData message
"""

from styler_rest_framework.pubsub.publishers.messages import Message


class SaveDataMessage(Message):
    """Generate SaveData message body"""

    def __init__(self, table, obj_type, key, data):
        self.name = "SaveData"
        self.arg = {"table": table, "obj_type": obj_type, "key": key, "data": data}



""" Middleware to handle exceptions
"""
import logging

from fastapi import FastAPI, Request, HTTPException
from fastapi.responses import JSONResponse


def add_exception_middleware(
    app: FastAPI,
    generic_message="An error has occurred",
    status_code=500,
    error_handler=None,
):
    """Generate a middleware that logs unexpected exceptions
    and returns a JSON response.
    Exceptions of type HTTPException won't be handled as they should be
    expected exceptions.

    Args:
        app: FastAPI application
        generic_message: The message that will be send as an error
        status_code: The HTTP status code (default = 500)
        error_handler: Callable(request, exception)
    """

    @app.middleware("http")
    async def handle_exception(request: Request, call_next):
        try:
            response = await call_next(request)
            return response
        except HTTPException:
            raise
        except Exception as ex:
            message = str(ex)
            if error_handler:
                error_handler(request, ex)
            logging.exception("Error: %s", message)
            return JSONResponse(
                status_code=status_code,
                content={"error": generic_message},
            )



""" Module for validations (i18n enabled)
"""

from styler_rest_framework import message
from styler_validation import ValidatorMixin
from styler_validation import validators  # NOQA


class I18nValidatorMixin(ValidatorMixin):
    """Translates validation messages"""

    def is_valid(self, locale="ja", **kwargs):
        result, errors = super().is_valid(**kwargs)
        if result:
            return result, errors
        localized_errors = {}
        for key, msg in errors.items():
            if len(msg) == 2:
                localized_errors[key] = message.get(
                    f"validation.{msg[0]}", locale=locale, value=msg[1]
                )
            else:
                localized_errors[key] = message.get(
                    f"validation.{msg[0]}", locale=locale
                )
        return result, localized_errors



from styler_rest_framework.pubsub.publishers.messages.logme_message import (
    LogMeMessage,
)
from styler_rest_framework.pubsub.publishers import publisher_for_message
from styler_rest_framework.config import defaults


logme = publisher_for_message(LogMeMessage, defaults.LOGME_TOPIC)



""" Configures logging with stackdriver
"""

import logging

import google.cloud.logging


def setup_logging(level=logging.INFO):  # pragma: no coverage
    """Setup logging"""
    client = google.cloud.logging.Client()
    excluded_logger_defaults = ("google.cloud", "google.auth", "google_auth_httplib2")

    # don't propagate excluded loggers (and don't send them to stderr either)
    for logger_name in excluded_logger_defaults:
        logging.getLogger(logger_name).propagate = False
        logging.getLogger(logger_name).addHandler(logging.NullHandler())

    handler = client.get_default_handler()
    handler.setLevel(level)
    root = logging.getLogger()
    root.addHandler(handler)
    root.setLevel(level)



""" Define helper methods to reduce the boilerplate code for a server
"""

import logging

from styler_middleware import handle_exceptions, handle_invalid_json
from styler_rest_framework.config import defaults
from styler_rest_framework.logging import setup_logging
from styler_rest_framework.logging.error_reporting import google_error_reporting_handler


def api_error_reporting_handler(service=None):  # pragma: no coverage
    try:
        service_name = service or defaults.ERROR_HANDLER_SERVICE
        from google.cloud import error_reporting

        handler = google_error_reporting_handler(service=service_name)

        def error_handler(request, exc):
            http_context = error_reporting.HTTPContext(
                method=request.method, url=request.path
            )
            handler(http_context=http_context)

        return error_handler
    except Exception:
        logging.warning("Could not find start api error reporting")
        return None


def add_middlewares(
    app,
    error_handler=None,
    service=None,
    handle_exceptions_args=None,
    handle_invalid_json_args=None,
):
    """Append default middlewares"""
    handle_exceptions_args = handle_exceptions_args or {}
    handle_invalid_json_args = handle_invalid_json_args or {}
    error_handler = error_handler or api_error_reporting_handler(service=service)
    app.middlewares.extend(
        [
            handle_exceptions(error_handler=error_handler, **handle_exceptions_args),
            handle_invalid_json(**handle_invalid_json_args),
        ]
    )


def set_logging(level=logging.INFO):  # pragma: no coverage
    setup_logging(level)


def default_configuration(app):  # pragma: no coverage
    set_logging()
    add_middlewares(app)



"""Firestore config
"""
import os

if os.getenv('ENVIRONMENT') == 'development':
    from mockfirestore import MockFirestore
    CLIENT = MockFirestore()
else:
    from firebase_admin import firestore
    import firebase_admin
    firebase_admin.initialize_app()
    CLIENT = firestore.client()



""" Tests for responses
"""
from unittest.mock import patch
import json

from styler_rest_framework.api import responses
from fastapi import HTTPException
from styler_rest_framework.exceptions.business import (
    InternalError,
    PermissionDeniedError,
    ResourceNotFoundError,
    ValidationError,
    ConflictError
)
from styler_rest_framework.exceptions.services import (
    AuthenticationError,
    AuthorizationError,
    InternalServerError,
    InvalidDataError,
    NotFoundError,
    PaymentRequiredError,
    UnexpectedError,
    ConflictError as ServiceConflictError,
)
import pytest


class TestStandard:
    """ Tests for method standard
    """
    def test_include_all_standard_codes(self):
        result = responses.standard()

        assert all([code in result for code in (400, 401, 403, 404, 409)])

    def test_include_none_if_invalid_code(self):
        result = responses.standard(486)

        assert all([code not in result for code in (400, 401, 403, 404, 409)])

    def test_include_only_desired_codes(self):
        result = responses.standard(401, 404)

        assert all([code in result for code in (401, 404)])
        assert all([code not in result for code in (400, 403, 409)])


class TestResponses:
    """ Tests for response handlers
    """

    cases = [
        (responses.unauthorized, 401),
        (responses.forbidden, 403),
        (responses.not_found, 404),
        (responses.conflict, 409)
    ]

    @pytest.mark.parametrize('method,expected', cases)
    def test_raises_http_exception(self, method, expected):
        with pytest.raises(HTTPException) as exp:
            method()

        assert exp.value.status_code == expected

    def test_bad_request(self):
        with pytest.raises(HTTPException) as exp:
            responses.bad_request({'error': 'my error'})

        assert exp.value.status_code == 400
        assert exp.value.detail == {
            'code': 'validation_error',
            'reason': {'error': 'my error'}
        }

    def test_payment_required(self):
        with pytest.raises(HTTPException) as exp:
            responses.payment_required(
                code='card_declined', reason='something bad')

        assert exp.value.status_code == 402
        assert exp.value.detail == {
            'code': 'card_declined',
            'reason': 'something bad'
        }

    @patch('logging.exception')
    def test_internal_server_error(self, mock_log):
        with pytest.raises(HTTPException) as exp:
            responses.internal_server_error(ValueError('some msg'))

        assert exp.value.status_code == 500
        mock_log.assert_called_once()


class TestHandleBusinessErrors:
    """ Tests for handle_business_errors
    """
    cases = [
        (InternalError, None, responses.internal_server_error),
        (ValidationError, {'error': 'something'}, responses.bad_request),
        (PermissionDeniedError, None, responses.forbidden),
        (ResourceNotFoundError, None, responses.not_found),
        (ConflictError, 'conflict', responses.conflict),
    ]

    @pytest.mark.parametrize('exp_type,args,expected', cases)
    def test_handle_errors(self, exp_type, args, expected):
        with patch(
            f'styler_rest_framework.api.responses.{expected.__name__}'
        ) as mocked_method:
            if args:
                responses.handle_business_errors(exp_type(args))
            else:
                responses.handle_business_errors(exp_type())

        mocked_method.assert_called_once()


class TestHandleServiceErrors:
    """ Tests for handle_service_errors
    """
    cases = [
        (AuthenticationError, 401, '', responses.unauthorized),
        (AuthorizationError, 403, '', responses.forbidden),
        (InternalServerError, 500, '', responses.internal_server_error),
        (
            InvalidDataError,
            400,
            {
                'code': 'validation_error',
                'reason': {'field': 'error'}
            },
            responses.bad_request
        ),
        (NotFoundError, 404, '', responses.not_found),
        (
            PaymentRequiredError,
            402,
            {'code': 'card_declined', 'reason': 'something bad'},
            responses.payment_required
        ),
        (UnexpectedError, 500, '', responses.internal_server_error),
        (ServiceConflictError, 409, '', responses.conflict),
    ]

    @pytest.mark.parametrize('exp_type,status,body,expected', cases)
    def test_handle_errors(self, exp_type, status, body, expected):
        with patch(
            f'styler_rest_framework.api.responses.{expected.__name__}'
        ) as mocked_method:
            response_text = json.dumps(body)
            responses.handle_service_errors(
                exp_type(status=status, response_text=response_text))

        mocked_method.assert_called_once()



from styler_rest_framework.pubsub.publishers.messages.send_mail_message import (
    SendMailMessage,
)
from styler_rest_framework.pubsub.publishers import publisher_for_message
from styler_rest_framework.config import defaults


send_email = publisher_for_message(SendMailMessage, defaults.MAILER_TOPIC)



from unittest.mock import Mock
from styler_rest_framework.crud.base import CRUDBase


class TestCRUDbase:
    """ Test CRUD base
    """
    def test_get(self):
        """ It should call related function
        """
        db = Mock()
        model = Mock()

        base = CRUDBase(model)

        base.get(db, 'test_id')

        db.query.assert_called_once()

    def test_get_multi(self):
        """ It should call related function
        """
        db = Mock()
        model = Mock()

        base = CRUDBase(model)

        base.get_multi(db)

        db.query.assert_called_once()

    def test_create(self):
        """ It should call related function
        """
        db = Mock()
        model = Mock()

        base = CRUDBase(model)

        base.create(db, obj_in={'key': 'value'})

        db.add.assert_called_once()
        db.commit.assert_called_once()
        db.refresh.assert_called_once()

    def test_update(self):
        """ It should call related function
        """
        db = Mock()
        model = Mock()

        base = CRUDBase(model)

        base.update(db, db_obj=Mock(), obj_in={'key': 'value'})

        db.add.assert_called_once()
        db.commit.assert_called_once()
        db.refresh.assert_called_once()

    def test_remove(self):
        """ It should call related function
        """
        db = Mock()
        model = Mock()

        base = CRUDBase(model)

        base.remove(db, id='test_id')

        db.query.assert_called_once()
        db.delete.assert_called_once()
        db.commit.assert_called_once()



""" Tests for the SaveData message
"""

import json

from styler_rest_framework.pubsub.publishers.messages.save_data_message \
    import (
        SaveDataMessage
    )


class TestSaveDataMessage:
    """ Tests for SaveData message
    """
    def test_generate_message(self):
        msg = SaveDataMessage(
            'my-tbl',
            'module#function',
            'some#key#123456789',
            {'aaa': 123, 'bbb': 333}
        )
        expected_keys_in_message = {
            'table', 'obj_type', 'key', 'data'
        }

        message_data = msg.encoded()
        decoded = json.loads(str(message_data, encoding='utf-8'))

        assert 'name' in decoded
        assert decoded['name'] == 'SaveData'
        assert 'arg' in decoded
        assert all([k in decoded['arg'] for k in expected_keys_in_message])



""" Handler for pubsub
"""
from concurrent import futures
import logging

from google.cloud import pubsub_v1


def get_callback(data):
    """Wrap message data in the context of the callback function."""

    def callback(api_future):
        try:
            api_future.result()
        except Exception:
            logging.error(
                "A problem occurred when publishing %s: %s\n",
                data,
                api_future.exception(),
            )
            raise

    return callback


def publish_message(topic_name, data):  # pragma: no coverage
    """Publishes a message to a Pub/Sub topic."""
    # Initialize a Publisher client.
    client = pubsub_v1.PublisherClient()
    metadata = {"version": "1"}

    # When you publish a message, the client returns a future.
    api_future = client.publish(topic_name, data=data, **metadata)
    api_future.add_done_callback(get_callback(data))

    futures.wait([api_future], return_when=futures.ALL_COMPLETED)



"""Tests for logme
"""
from unittest.mock import patch

from styler_rest_framework.helpers import logme


class TestSendMail:
    """Tests for send_mail
    """
    @patch('styler_rest_framework.pubsub.publishers.publish_message')
    def test_send_mail(self, mocked_message):
        logme.logme(
            'test_dataset',
            'test_table',
            [{"field": "value"}, {"field": "value2"}]
        )

        mocked_message.assert_called_once()


""" HTTP error schemas
"""
from enum import Enum
from typing import Dict

from pydantic import BaseModel


class PaymentErrorCode(str, Enum):
    card_declined = "card_declined"
    contact_support = "contact_support"
    try_again = "try_again"
    validation_error = "validation_error"
    unsupported_card = "unsupported_card"


class HTTP400Error(BaseModel):
    """Raised when handling HTTPExceptions with code 400"""

    code: str
    reason: Dict[str, str]


class HTTP401Error(BaseModel):
    """Returned by istio"""


class HTTP402Error(BaseModel):
    """Returned by istio"""

    code: PaymentErrorCode
    reason: str


class HTTP403Error(BaseModel):
    """Raised when handling HTTPExceptions with code 403"""


class HTTP404Error(BaseModel):
    """Raised when handling HTTPExceptions with code 404"""


class HTTP409Error(BaseModel):
    """Raised when handling HTTPExceptions with code 409"""



""" Messages templates
"""

from styler_rest_framework.pubsub.publishers.messages import Message
from styler_rest_framework.config import defaults


class SendMailMessage(Message):
    """Generate SendMail message body"""

    def __init__(
        self,
        receiver,
        title,
        body,
        cc=None,
        sender_email=None,
        sender_name=None,
        email_type=None,
    ):
        self.name = "SendMail"
        self.arg = {}
        self._fill_args(
            sender_email, receiver, title, body, cc, sender_name, email_type
        )

    def _fill_args(
        self, sender_email, receiver, title, body, cc, sender_name, email_type
    ):
        """Map fields from email"""
        self.arg = {
            "sender": [
                sender_email or defaults.EMAIL_SENDER,
                sender_name or defaults.EMAIL_SENDER_NAME,
            ],
            "receiver": [receiver],
            "cc": cc or [],
            "title": title,
            "body": body,
            "type": email_type or defaults.EMAIL_TYPE,
        }



""" Module for controllers
"""

import json
import logging

from aiohttp import web
from styler_rest_framework.exceptions.services import (
    AuthenticationError,
    AuthorizationError,
    InternalServerError,
    InvalidDataError,
    NotFoundError,
    PaymentRequiredError,
    UnexpectedError,
)
from styler_rest_framework.exceptions.business import (
    InternalError,
    PermissionDeniedError,
    ResourceNotFoundError,
    ValidationError,
    ConflictError,
)
from styler_rest_framework.api.request_scope import RequestScope


class BaseController:
    """Base class for controllers"""

    def get_identity(self, request):
        """Get the identity of the logged user"""
        try:
            return RequestScope.from_request(request)
        except ValueError:
            self.bad_request({"jwt": "invalid"})

    def resp_for_creation(self, new_id):
        """Default response for creation methods"""
        return web.json_response({"id": new_id}, status=201)

    def response_ok(self, data):
        """Default response for 200 OK"""
        return web.json_response(data)

    def resp_for_listing(self, paginator, items):
        """Default response for listings with pagination"""
        return web.json_response({"items": items, "pagination": paginator.get_info()})

    def payment_required(self, code, reason):
        """Default response for Error code 402"""
        error = {"code": code, "reason": reason}
        logging.warning("Payment required error: %s", error)
        raise web.HTTPPaymentRequired(
            text=json.dumps(error), content_type="application/json"
        )

    def bad_request(self, errors, code="validation_error"):
        """Default response for Error code 400"""
        error = {"code": code, "reason": errors}
        logging.warning("Bad request error: %s", error)
        raise web.HTTPBadRequest(
            text=json.dumps(error), content_type="application/json"
        )

    def not_found(self, msg=""):
        """Default response for Error code 404"""
        logging.warning("Not found error error: %s", msg)
        raise web.HTTPNotFound(content_type="application/json")

    def unauthorized(self, msg=""):
        """Default response for Error code 401"""
        logging.warning("Unauthorized error: %s", msg)
        raise web.HTTPUnauthorized(content_type="application/json")

    def conflict(self, msg=""):
        """Default response for Error code 409"""
        logging.warning("Conflict error: %s", msg)
        raise web.HTTPConflict(content_type="application/json")

    def forbidden(self, msg=""):
        """Default response for Error code 403"""
        logging.warning("Forbidden error: %s", msg)
        raise web.HTTPForbidden(content_type="application/json")

    def internal_server_error(self, exception, msg=""):
        """Default response for Error code 500"""
        logging.exception("Internal Server error: %s", str(exception))
        raise web.HTTPInternalServerError(content_type="application/json")

    def handle_service_errors(self, ex):
        """Default method for handling service related errors"""
        if isinstance(ex, InvalidDataError):
            self.bad_request(ex.json_body().get("reason"))
        elif isinstance(ex, AuthenticationError):
            self.unauthorized()
        elif isinstance(ex, PaymentRequiredError):
            body = ex.json_body()
            self.payment_required(body.get("code"), body.get("reason"))
        elif isinstance(ex, AuthorizationError):
            self.forbidden()
        elif isinstance(ex, NotFoundError):
            self.not_found()
        elif isinstance(ex, InternalServerError):
            self.internal_server_error(exception=ex)
        elif isinstance(ex, UnexpectedError):
            self.internal_server_error(exception=ex)
        else:
            raise ex

    def handle_business_errors(self, ex):
        """Default method for handling business related errors"""
        if isinstance(ex, ValidationError):
            self.bad_request(ex.errors)
        elif isinstance(ex, PermissionDeniedError):
            self.forbidden()
        elif isinstance(ex, ResourceNotFoundError):
            self.not_found()
        elif isinstance(ex, ConflictError):
            self.conflict()
        elif isinstance(ex, InternalError):
            self.internal_server_error(exception=ex)
        else:
            raise ex



""" Session management tools
"""
from typing import Generator

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker


_engine = None
_session_local = None


def get_session(database_uri: str) -> Generator:
    global _engine
    global _session_local
    _engine = _engine or create_engine(database_uri, pool_pre_ping=True)
    _session_local = _session_local or sessionmaker(
        autocommit=False, autoflush=False, bind=_engine
    )
    try:
        session_instance = _session_local()
        yield session_instance
    finally:
        session_instance.close()



""" Tests for server defaults
"""

from unittest.mock import Mock

from styler_rest_framework.helpers import aiohttp_defaults


class TestAddMiddlewares:
    def test_add_middlewares(self):
        app = Mock()
        app.middlewares = []

        aiohttp_defaults.add_middlewares(app)

        assert len(app.middlewares) == 2
        assert callable(app.middlewares[0])
        assert callable(app.middlewares[1])

    def test_add_middlewares_with_custom_error_handler(self):
        app = Mock()
        app.middlewares = []
        error_handler = Mock()

        aiohttp_defaults.add_middlewares(app, error_handler=error_handler)

        assert len(app.middlewares) == 2
        assert callable(app.middlewares[0])
        assert callable(app.middlewares[1])



from functools import wraps
import json
import time

from styler_rest_framework.api.request_scope import Identity

from styler_rest_framework.pubsub.publishers.messages.save_data_message import (
    SaveDataMessage,
)
from styler_rest_framework.pubsub.publishers import publisher_for_message
from styler_rest_framework.config import defaults


handler = publisher_for_message(SaveDataMessage, defaults.TOPIC_NAME)

_TABLE = f"userEvents-{defaults.ENVIRONMENT}"


def track(func):
    """Track and log all inputs and outputs"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        data = {**extract_args(args), **extract_named_args(kwargs)}
        try:
            result = func(*args, **kwargs)
            data["return"] = json.dumps(result, skipkeys=True)
        finally:
            obj_type = ".".join([func.__module__, func.__name__])
            args_dict = {k: v for k, v in zip(range(len(args)), args)}
            key = "#".join(
                [str(get_user_id({**args_dict, **kwargs})), obj_type, str(time.time())]
            )
            handler(_TABLE, obj_type, key, data)
        return result

    return wrapper


def extract_args(args):
    """Extract the args"""
    inputs = {}
    for arg in args:
        try:
            if "__dict__" in dir(arg):
                value = json.dumps(arg.__dict__)
            elif type(arg) in (int, str, float, bool):
                value = arg
            else:
                value = json.dumps(arg)
        except Exception:
            value = repr(arg)
        inputs[len(inputs.keys())] = value
    return inputs


def extract_named_args(kwargs):
    """Extract the kwargs"""
    inputs = {}
    if not kwargs:
        return inputs
    for k, kwarg in kwargs.items():
        try:
            if "__dict__" in dir(kwarg):
                value = json.dumps(kwarg.__dict__)
            elif type(kwarg) in (int, str, float, bool):
                value = kwarg
            else:
                value = json.dumps(kwarg)
        except Exception:
            value = repr(kwarg)
        inputs[k] = value
    return inputs


def get_user_id(data):
    """Extract the user_id from the identity"""
    if not data:
        return None
    for k, v in data.items():
        if isinstance(v, Identity):
            return v.user_id()
    return None



""" Errors raised by business classes
"""


class BusinessError(Exception):  # pragma: no coverage
    """Error raised from business classes"""


class InternalError(BusinessError):  # pragma: no coverage
    """Error raised when an unexpected error occurs"""


class PermissionDeniedError(BusinessError):  # pragma: no coverage
    """Error raised when the user has no access to the resource"""


class ResourceNotFoundError(BusinessError):  # pragma: no coverage
    """Error raised when the required resource was not found"""


class ValidationError(BusinessError):  # pragma: no coverage
    """Error raised when there is a validation error"""

    def __init__(self, errors):
        super().__init__()
        self.errors = errors


class ConflictError(BusinessError):
    """Error raised when request could not be completed due to a conflict
    with the current state of the resource
    """

    def __init__(self, msg):
        super().__init__()
        self.msg = msg



""" Message template for logme
"""

from styler_rest_framework.pubsub.publishers.messages import Message
from styler_rest_framework.config import defaults

class LogMeMessage(Message):
    """ Generates LogMe message body. """

    def __init__(self, dataset, table, rows):
        self.name = "LogMe"
        self.arg = {}
        self._fill_args(dataset, table, rows)

    def _fill_args(self, dataset, table, rows):
        """ Map fields """
        self.arg = {
            "dataset": dataset,
            "table": table,
            "rows": rows
        }


""" Middleware to handle exceptions
"""
import logging

try:
    from fastapi import FastAPI, Request, HTTPException
    from fastapi.responses import JSONResponse
except Exception:
    logging.exception(f'Missing libraries: pipenv install fastapi')

from styler_rest_framework.middlewares.fastapi.jwt_validator import validate
from styler_rest_framework.config import defaults


def add_auth_middleware(
    app: FastAPI,
    env: str,
    jwks_url: str = None
):
    if not jwks_url:
        jwks_url = defaults.JWKS_URL

    @app.middleware("http")
    async def validate_jwt(request: Request, call_next):
        try:
            jwt_token = request.headers.get('Authorization')
            if not jwt_token:
                return JSONResponse(status_code=401, content={'error': 'Missing JWT token'})
            jwt_data = validate(jwt_token.split()[1], env, jwks_url)
            if not jwt_data:
                return JSONResponse(status_code=401, content={'error': 'Invalid JWT token'})

            response = await call_next(request)
            return response
        except HTTPException:
            raise



"""Tests for mailer
"""
from unittest.mock import patch

from styler_rest_framework.helpers import mailer


class TestSendMail:
    """Tests for send_mail
    """
    @patch('styler_rest_framework.pubsub.publishers.publish_message')
    def test_send_mail(self, mocked_message):
        mailer.send_email('some@mail.com', 'Some title', 'a body')

        mocked_message.assert_called_once()



""" Services module
"""

import logging

from aiohttp import ClientSession
from styler_rest_framework.exceptions.services import (
    AuthenticationError,
    AuthorizationError,
    InternalServerError,
    InvalidDataError,
    NotFoundError,
    PaymentRequiredError,
    ServiceError,
    UnexpectedError,
    ConflictError,
)


class HTTPHandler:
    """Handles HTTP operations"""

    def __init__(self, identity=None, retry_on=None, headers=None):
        if retry_on is None:
            self.retry_on = [503]
        else:
            self.retry_on = retry_on
        self.headers = {}
        if headers is None:
            headers = {}
        if identity:
            self.identity = identity
            self.headers = {
                "Authorization": f"Bearer {identity.token()}",
                **{"Accept-Language": identity.localization()},
                **identity.trace_header()
            }
        self.headers = {
            **self.headers,
            **headers
        }

    async def post(self, url, params, error_handlers=None, retry=3, **kwargs):
        retry_on = kwargs.get("retry_on") or self.retry_on
        headers = self._prepare_headers(kwargs.get("headers", {}))
        try:
            async with ClientSession() as session:
                async with session.post(
                    url, headers=headers, json=params, ssl=True
                ) as resp:
                    if resp.status not in [200, 201]:
                        if error_handlers and resp.status in error_handlers:
                            await error_handlers[resp.status](resp)
                        await self._handle_http_errors(resp)

                    return await resp.json()
        except ServiceError as ex:
            # Retry
            if retry > 0 and ex.status in retry_on:
                return await self.post(
                    url, params, error_handlers=None, retry=retry - 1, **kwargs
                )
            else:
                raise

    async def patch(self, url, params, error_handlers=None, retry=3, **kwargs):
        retry_on = kwargs.get("retry_on") or self.retry_on
        headers = self._prepare_headers(kwargs.get("headers", {}))

        try:
            async with ClientSession() as session:
                async with session.patch(
                    url, headers=headers, json=params, ssl=True
                ) as resp:
                    if resp.status not in [200]:
                        if error_handlers and resp.status in error_handlers:
                            await error_handlers[resp.status](resp)
                        await self._handle_http_errors(resp)

                    return await resp.json()
        except ServiceError as ex:
            # Retry
            if retry > 0 and ex.status in retry_on:
                return await self.patch(
                    url, params, error_handlers=None, retry=retry - 1, **kwargs
                )
            else:
                raise

    async def put(self, url, params, error_handlers=None, retry=3, **kwargs):
        retry_on = kwargs.get("retry_on") or self.retry_on
        headers = self._prepare_headers(kwargs.get("headers", {}))
        try:
            async with ClientSession() as session:
                async with session.put(
                    url, headers=headers, json=params, ssl=True
                ) as resp:
                    if resp.status not in [200]:
                        if error_handlers and resp.status in error_handlers:
                            await error_handlers[resp.status](resp)
                        await self._handle_http_errors(resp)

                    return await resp.json()
        except ServiceError as ex:
            # Retry
            if retry > 0 and ex.status in retry_on:
                return await self.put(
                    url, params, error_handlers=None, retry=retry - 1, **kwargs
                )
            else:
                raise

    async def delete(self, url, error_handlers=None, retry=3, **kwargs):
        retry_on = kwargs.get("retry_on") or self.retry_on
        headers = self._prepare_headers(kwargs.get("headers", {}))
        try:
            async with ClientSession() as session:
                async with session.delete(url, headers=headers, ssl=True) as resp:
                    if resp.status not in [200]:
                        if error_handlers and resp.status in error_handlers:
                            await error_handlers[resp.status](resp)
                        await self._handle_http_errors(resp)

                    return await resp.json()
        except ServiceError as ex:
            # Retry
            if retry > 0 and ex.status in retry_on:
                return await self.delete(
                    url, error_handlers=None, retry=retry - 1, **kwargs
                )
            else:
                raise

    async def get(self, url, error_handlers=None, retry=3, **kwargs):
        retry_on = kwargs.get("retry_on") or self.retry_on
        headers = self._prepare_headers(kwargs.get("headers", {}))
        try:
            async with ClientSession() as session:
                async with session.get(url, headers=headers, ssl=True) as resp:
                    if resp.status != 200:
                        if error_handlers and resp.status in error_handlers:
                            await error_handlers[resp.status](resp)
                        await self._handle_http_errors(resp)

                    return await resp.json()
        except ServiceError as ex:
            # Retry
            if retry > 0 and ex.status in retry_on:
                return await self.get(
                    url, error_handlers=None, retry=retry - 1, **kwargs
                )
            else:
                raise

    def _prepare_headers(self, overrides):
        headers = self.headers
        headers.update(overrides)
        return headers

    async def _handle_http_errors(self, resp):
        """Handles HTTP errors"""
        response_text = await resp.text()
        if resp.status == 400:
            raise InvalidDataError(resp.status, response_text)
        elif resp.status == 401:
            raise AuthenticationError(resp.status, response_text)
        elif resp.status == 402:
            raise PaymentRequiredError(resp.status, response_text)
        elif resp.status == 403:
            raise AuthorizationError(resp.status, response_text)
        elif resp.status == 404:
            raise NotFoundError(resp.status, response_text)
        elif resp.status == 409:
            raise ConflictError(resp.status, response_text)
        elif resp.status == 500:
            raise InternalServerError(resp.status, response_text)
        else:
            logging.error("Unexpected response code: %s", resp.status)
            raise UnexpectedError(resp.status, response_text)
